
from abc import *
import time, logging, json
from .Prompt import Prompt
from .GPTRequest import RawChatGPTQueue, GPTRequest
class AbstractLogic(metaclass=ABCMeta):
    def __init__(self, model_card, add_welcome_message=True, unique_id="SEOWOO_V2", version=2, creator="SW",
                 gpt_queue=None):
        logging.basicConfig(level=logging.INFO)
        self.unique_id = unique_id
        self.version = version
        self.creator = creator

        logging.info("[%s] %s에 의해 제작된 채팅 처리 로직 (VERSION %d)" % (unique_id, creator, version))
        logging.info("CHARACTER CARD ID : %s" % model_card.name)

        self.model_card = model_card
        self.model_card_id = model_card.name

        logging.info("모델 카드")
        logging.info(self.model_card)
        self.messages = []
        self.gpt_queue = gpt_queue

        if add_welcome_message:
            self.messages.append(Prompt.from_dict(self.model_card.get_random_messages(self.model_card.messages_welcome)))

        if gpt_queue is None:
            self.gpt_queue = RawChatGPTQueue(num_workers=1)


        self.logs = {}
        self.log_index = 0


    def reset_messages(self):
        self.messages.clear()


    def request_gpt(self, request_dic):
        self.log_index = self.log_index + 1
        self.logs["request_"+str(self.log_index)] = request_dic

        print(json.dumps(request_dic,indent=4,ensure_ascii=False))

        request = GPTRequest(request_dic)
        self.gpt_queue.q.put(request)

        while True:
            time.sleep(0.1)
            if request.status == "FINISHED":
                response = request.response
                self.logs["response_" + str(self.log_index)] = response
                del request
                print(json.dumps(response, indent=4, ensure_ascii=False))
                return response

            if request.status == "ERROR":
                return None

    @abstractmethod
    def process(self, user_input_text):
        pass

    def register_assistant_text(self, assistant_response_text, hidden_input_text=""):
        self.messages.append(Prompt(role='assistant',content=assistant_response_text, hidden_content=hidden_input_text))
        return assistant_response_text, hidden_input_text

    def register_user_input_text(self, user_input_text, hidden_input_text=""):
        self.messages.append(Prompt(role='user',content=user_input_text,hidden_content=hidden_input_text))
        return user_input_text, hidden_input_text




import logging, json, traceback, time
from Config import Config

class GPTRequest:
    def __init__(self, params):
        if type(params) is not dict:
            assert 0

        self.status = "PENDING"
        self.response = None
        self.request = params

class LogicTask:
    def __init__(self, model, messages):
        self.model = model
        self.messages = messages
        self.timestamp = time.time()

import queue, threading
from openai import OpenAI
class RawChatGPTQueue:
    DEBUG = False

    def __init__(self, num_workers):
        self.q = queue.Queue()
        self.num_workers = num_workers
        self.threads = []
        self.start_workers()
        self.client = OpenAI(
            api_key=Config.OPENAI_API_KEY,
            organization=Config.OPENAI_ORG
        )
    def query_gpt(self):
        while True:
            logging.debug("ChatGPTQueue %d개"%self.q.qsize())

            request = self.q.get()
            if request is None:
                break

            try:
                if request.status =="PENDING":
                    request.status = "REQUESTED"
                    dic = request.request
                    logging.debug("CHATGPT 요청")
                    logging.debug(json.dumps(dic,sort_keys=True, ensure_ascii=False, indent=4))
                    print(dic)

                    if dic.get('request_timeout',None) is not None:
                        dic['timeout'] = dic['request_timeout']
                        del dic['request_timeout']

                    response =  self.client.chat.completions.create(
                        **dic,
                    )
                    # for compatibility
                    response = {
                        'choices':[{
                            'message':{
                                'content':response.choices[0].message.content
                            }
                        }],
                        'model' : response.model,
                        'usage' : {
                            'completion_tokens' : response.usage.completion_tokens,
                            'prompt_tokens' : response.usage.prompt_tokens,
                            'total_tokens' : response.usage.total_tokens,
                        }
                    }

                    request.response = response
                    request.status = "FINISHED"

                    logging.debug("CHATGPT 회신")
                    logging.debug(response)

            except Exception as e:
                logging.error("CHATGPT 에러 발생 : " + str(e))
                logging.error(traceback.format_exc())
                request.status = "ERROR"

            self.q.task_done()

    def start_workers(self):
        threads = []
        for i in range(self.num_workers):
            t = threading.Thread(target=self.query_gpt)
            t.start()
            self.threads.append(t)
        return threads


    def stop_workers(self):
        # stop workers
        for i in self.threads:
            self.q.put(None)
        for t in self.threads:
            t.join()


import tiktoken

class Tokenizer:

    @classmethod
    def encoder(cls, backend):
        try:
            encoder = tiktoken.encoding_for_model(backend)
        except KeyError:
            print("Warning: model not found. Using cl100k_base encoding.")
            encoder = tiktoken.get_encoding("cl100k_base")
        if backend in {
            "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k-0613",
            "gpt-4-0314",
            "gpt-4-32k-0314",
            "gpt-4-0613",
            "gpt-4-32k-0613",
        }:
            tokens_per_message = 3
            tokens_per_name = 1
        elif backend == "gpt-3.5-turbo-0301":
            tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            tokens_per_name = -1  # if there's a name, the role is omitted
        elif "gpt-3.5-turbo" in backend:
            print("Warning: gpt-3.5-turbo may update over time. Returning num tokens assuming gpt-3.5-turbo-0613.")
            return cls.encoder(backend="gpt-3.5-turbo-0613")
        elif "gpt-4" in backend:
            print("Warning: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613.")
            return cls.encoder(backend="gpt-4-0613")
        else:
            raise NotImplementedError(
                f"""num_tokens_from_messages() is not implemented for model {backend}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
            )
        return encoder, tokens_per_message, tokens_per_name


    @classmethod
    def num_tokens_from_messages(cls, messages, backend="gpt-3.5-turbo-0613"):

        if type(messages) is dict:
            messages = [messages]

        """Return the number of tokens used by a list of messages."""
        encoder, tokens_per_message, tokens_per_name = cls.encoder(backend)

        num_tokens = 0
        for message in messages:
            num_tokens += tokens_per_message
            for key, value in message.items():
                num_tokens += len(encoder.encode(value))
                if key == "name":
                    num_tokens += tokens_per_name
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens



import json, logging
from .Tokenizer import Tokenizer
__all__ = ['Prompt','PromptSection','GPTPrompt']

class Prompt:
    def __init__(self, role, content, hidden_content="", unique_id=None, backend="gpt-3.5-turbo-0613"):

        self.role = role
        self.content = content
        if hidden_content == "":
            hidden_content = content
        self.hidden_content  = hidden_content
        self.backend = backend
        self.unique_id = unique_id

        self.dic = {'role': role, 'content': content, 'hidden' :hidden_content}

    @classmethod
    def from_dict(cls, dict):
        return cls(**dict)

    @property
    def token(self):
        return Tokenizer.num_tokens_from_messages(self.dic, backend=self.backend)

    def __str__(self):
        return json.dumps(self.dic, sort_keys=True, ensure_ascii=False, indent=4)

    def __dict__(self):
        return {'role': self.role, 'content': self.content}

    def truncate_sentences(self, limit):
        if self.token <= limit:
            return
        sentences = self.content.split(". ")
        while True:
            if self.token <= limit:
                break
            sentences.pop(0)
            self.content = ". ".join(sentences)

    def __getstate__(self):
        return {'role':self.role,'content':self.content,'hidden_content':self.hidden_content,
                'unique_id':self.unique_id,'backend':self.backend}

    def __setstate__(self, d):
        self.role = d['role']
        self.content = d['content']
        self.hidden_content = d['hidden_content']
        self.unique_id = d['unique_id']
        self.backend = d['backend']


class PromptSection:
    def __init__(self, section, prompt=None, min_token=None, max_token=None):
        if prompt is None:
            prompt = []
        if type(prompt) is not list:
            prompt = [prompt]

        self.section = section
        self.prompt = prompt

        if max_token is not None:
            self.truncate(max_token)

        self.max_token = max_token
        self.min_token = min_token

        if max_token is not None and min_token is not None:
            assert max_token >= min_token

    def __len__(self):
        return len(self.prompt)

    @property
    def token(self):
        token = 0
        for p in self.prompt:
            token = token + p.token

        # print("%s - total %d - token %d"%(self.section,len(self.prompt),token))
        return token

    @property
    def num(self):
        return len(self.prompt)



    def truncate(self, limit=None):
        if limit is None:
            limit = self.max_token
        truncated = []
        # print("PromptSection Truncate (총 %d개) %d -> %d"% (len(self.prompt),self.token,limit))
        while True:
            if self.token <= limit:
                break
            if self.min_token is not None:
                if self.token <= self.min_token:
                    #print("min_break")
                    break
            removed = self.prompt.pop(0)
            #print("removed : %s" % removed.content)
            truncated.append(removed)
            #print(truncated)

        if len(truncated) > 0:
            logging.debug("PromptSection %s 에서 오래된 항목 %d개 제거" % (self.section, len(truncated)))
        return PromptSection(section="truncated", prompt=truncated)

    def append(self, prompt: Prompt, auto_truncate: bool = True):
        self.prompt.append(prompt)

        if auto_truncate:
            if self.max_token is not None:
                self.truncate(self.max_token)

    def toJSON(self):
        return str(self)

    def toList(self):
        return json.loads(str(self))

    def toPromptList(self):
        arr = []
        for p in self.prompt:
            arr.append(p)
        return arr


    def __str__(self):
        arr = []
        for p in self.prompt:
            arr.append({'role': p.role, 'content': p.content})

        return json.dumps(arr, sort_keys=True, ensure_ascii=False, indent=4)


class GPTPrompt:
    def __init__(self):
        self.section_list = []

        self.prompt_important_order = ['character_system', 'few_shot_assistant', 'summary_system', 'dialogue',
                                       'pre_user_text_system', 'user_text', 'post_user_text_system']

    def put(self, section: PromptSection):
        self.section_list.append(section)

    @property
    def token(self):
        token = 0
        for section in self.section_list:
            token = token + section.token
        return token

    def truncate(self, token_limit):
        while True:
            if self.token <= token_limit:
                break

            have_to_reduce_token = self.token - token_limit
            logging.info("GPTPrompt Truncation : have to reduce %d tokens" % have_to_reduce_token)

            for to_truncate_section in self.prompt_important_order.reverse():
                prompt_section = None
                for prompt_section in self.section_list:
                    if prompt_section.section == to_truncate_section:
                        break
                if prompt_section is not None:
                    prompt_section = self.prompt_section_dic[to_truncate_section]
                    prompt_section.truncate(limit=prompt_section.token - have_to_reduce_token)

    def __str__(self):
        # print("SECTION_LIST")

        final_list = []
        for section in self.section_list:

            # print(section.section)
            # print(len(section.prompt))
            # print(id(section.prompt))

            for p in section.prompt:
                final_list.append({'role': p.role, 'content': p.content})

        return json.dumps(final_list, sort_keys=True, ensure_ascii=False, indent=4)




from setuptools import setup, find_packages

setup(
    name='angelsupport',
    version='0.0.5',
    description='Angel Support Program',
    author='Anthony',
    author_email='webmaster@ailike.me',
    url='https://ailike.me',
    install_requires=['openai', 'tiktoken'],
    packages=find_packages(exclude=[]),
    keywords=[],
    python_requires='>=3.6',
    package_data={},
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



import logging, os, random, json, base64

from .Tokenizer import Tokenizer

class ModelCardReader:
    @classmethod
    def read_model_from_path(cls, path, override_dic=None):
        print(path)
        with open(path, 'r',encoding="utf-8") as file:
            dic = json.load(file)
            if override_dic is not None:
                for key, value in override_dic.items():
                    logging.info("MODELCARD_READER OVERRIDE PARAMETER "+key)
                    dic[key] = value
            return dic

    @classmethod
    def read_from_server(cls, model_card_id, override_dic=None):
        assert 0
        return character_dic


class ModelGenerator:
    def __init__(self, model_folder_path, override_dic={}):
        self.model_cards = {}
        for model in os.listdir(model_folder_path):
            card = None
            if model.lower().endswith(".json"):
                model_json_path = os.path.join(model_folder_path, model)
                config = ModelCardReader.read_model_from_path(path=model_json_path, override_dic=override_dic)
                if 'version' in config.keys():
                    if float(config['version']) == 2:
                        card = ModelCardV2(**config)

                if card is None:
                    logging.error("모델 카드 버전을 확인할 수 없습니다.")
                    raise Exception
                self.model_cards[card.name] = card

    def model(self, model_name):
        if model_name in self.model_cards.keys():
            return self.model_cards[model_name]
        else:
            model_info = ModelCardReader.read_from_server(model_name)
            model = ModelCardV2(**model_info)
            self.model_cards[model_name] = model
            return model



class ModelCardV2:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def get_random_messages(self, messages):
        return random.choice(messages)

    def num_tokens_from_messages(self, messages):
        return Tokenizer.num_tokens_from_messages(messages=messages, backend=self.model_backend)
